from rest_framework.pagination import PageNumberPagination as _PageNumberPagination, \
    LimitOffsetPagination as _LimitOffsetPagination, CursorPagination as _CursorPagination

from djackal.settings import djackal_settings


class PageNumberPagination(_PageNumberPagination):
    page_size = djackal_settings.PAGE_SIZE
    page_size_query_param = 'page_size'
    max_page_size = djackal_settings.MAX_PAGE_SIZE


class LimitOffsetPagination(_LimitOffsetPagination):
    pass


class CursorPagination(_CursorPagination):
    pass



import json

from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import TextField


def dumps(value):
    return json.dumps(
        value,
        cls=DjangoJSONEncoder,
        sort_keys=True,
        indent=2,
        separators=(',', ': ')
    )


class JSONField(TextField):
    def __init__(self, *args, **kwargs):
        self.serializer = kwargs.pop('serializer', dumps)
        self.deserializer = kwargs.pop('deserializer', json.loads)
        super(JSONField, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(JSONField, self).deconstruct()
        kwargs['serializer'] = self.serializer
        kwargs['deserializer'] = self.deserializer
        return name, path, args, kwargs

    def to_python(self, value):
        if value == "":
            return None
        try:
            if isinstance(value, str):
                return self.deserializer(value)
            elif isinstance(value, bytes):
                return self.deserializer(value.decode('utf8'))
        except ValueError:
            pass
        return value

    def get_prep_value(self, value):
        if value == "":
            return None
        if isinstance(value, (dict, list)):
            return self.serializer(value)
        return super(JSONField, self).get_prep_value(value)

    def from_db_value(self, value, *args, **kwargs):
        return self.to_python(value)

    def get_default(self):
        if self.has_default():
            return self.default() if callable(self.default) else self.default
        return ""

    def get_db_prep_save(self, value, *args, **kwargs):
        if value == "":
            return None
        if isinstance(value, (dict, list)):
            return self.serializer(value)
        else:
            return super(JSONField,
                         self).get_db_prep_save(value, *args, **kwargs)

    def value_from_object(self, obj):
        value = super(JSONField, self).value_from_object(obj)
        if self.null and value is None:
            return None
        return self.serializer(value)



import collections


def value_mapper(a_dict, b_dict):
    """
    a_dict = {
        'key1': 'a_value1',
        'key2': 'a_value2'
    }

    b_dict = {
        'key1': 'b_value1',
        'key2': 'b_value2'
    }

    >>> value_mapper(a_dict, b_dict)
    {
        'a_value1': 'b_value1',
        'a_value2': 'b_value2'
    }
    """

    return {a_value: b_dict.get(a_key) for a_key, a_value in a_dict.items()}


def isiter(arg):
    return isinstance(arg, collections.Iterable) and not isinstance(arg, str)



from django.apps import apps
from django.db.models import Q, Model
from django.shortcuts import _get_queryset

from djackal.exceptions import NotFound
from djackal.settings import djackal_settings


def get_object_or_None(klass, *args, **kwargs):
    queryset = _get_queryset(klass)
    try:
        return queryset.get(*args, **kwargs)
    except queryset.model.DoesNotExist:
        return None


def get_object_or(klass, this=None, *args, **kwargs):
    return get_object_or_None(klass, *args, **kwargs) or this


def get_object_or_404(model, **fields):
    obj = get_object_or_None(model, **fields)
    if obj is None:
        raise NotFound(model=model)
    return obj


def model_update(instance, **fields):
    for key, value in fields.items():
        setattr(instance, key, value)

    instance.save()
    return instance


def get_model(label, *args, **kwargs):
    if djackal_settings.SINGLE_APP:
        if djackal_settings.SINGLE_APP_NAME and len(label.split('.')) == 1:
            return apps.get_model('{}.{}'.format(djackal_settings.SINGLE_APP_NAME, label))
    return apps.get_model(label, *args, **kwargs)


def gen_q(value, *filter_keywords):
    q_object = Q()
    for q in filter_keywords:
        q_object |= Q(**{q: value})
    return q_object


def auto_f_key(**kwargs):
    result = dict()

    for key, value in kwargs.items():
        if type(value) is int:
            result['{}_id'.format(key)] = value
        elif isinstance(value, Model):
            result[key] = value
        else:
            raise ValueError('unknown type: {}'.format(type(value)))
    return result



from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Storage',
            fields=[
                ('key', models.CharField(max_length=150, primary_key=True, serialize=False)),
                ('value', models.TextField()),
            ],
        ),
    ]



from rest_framework import serializers

from djackal.views.base import DjackalAPIView
from . import mixins
from ..serializers import BaseModelSerializer

__all__ = [
    'ListAPIView',
    'CreateAPIView',
    'ListCreateAPIView',
    'DetailAPIView',
    'UpdateAPIView',
    'DestroyAPIView',
    'DetailUpdateAPIView',
    'DetailDestroyAPIView',
    'UpdateDestroyAPIView',
    'DetailUpdateDestroyAPIView',
    'LabelValueListAPIView',
]


class ListAPIView(DjackalAPIView, mixins.ListViewMixin):
    def get(self, request, **kwargs):
        return self.list(request, **kwargs)


class CreateAPIView(DjackalAPIView, mixins.CreateViewMixin):
    def post(self, request, **kwargs):
        return self.create(request, **kwargs)


class ListCreateAPIView(DjackalAPIView, mixins.ListViewMixin, mixins.CreateViewMixin):
    def get(self, request, **kwargs):
        return self.list(request, **kwargs)

    def post(self, request, **kwargs):
        return self.create(request, **kwargs)


class DetailAPIView(DjackalAPIView, mixins.DetailViewMixin):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)


class UpdateAPIView(DjackalAPIView, mixins.UpdateViewMixin):
    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)


class DestroyAPIView(DjackalAPIView, mixins.DestroyViewMixin):
    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class DetailUpdateAPIView(DjackalAPIView, mixins.DetailViewMixin, mixins.UpdateViewMixin):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)

    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)


class DetailDestroyAPIView(DjackalAPIView, mixins.DetailViewMixin, mixins.DestroyViewMixin):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)

    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class UpdateDestroyAPIView(DjackalAPIView, mixins.UpdateViewMixin, mixins.DestroyViewMixin):
    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)

    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class DetailUpdateDestroyAPIView(DjackalAPIView, mixins.DetailViewMixin, mixins.UpdateViewMixin,
                                 mixins.DestroyViewMixin):
    def get(self, request, **kwargs):
        return self.detail(request, **kwargs)

    def patch(self, request, **kwargs):
        return self.update(request, **kwargs)

    def delete(self, request, **kwargs):
        return self.destroy(request, **kwargs)


class LabelValueListAPIView(DjackalAPIView):
    label_field = 'name'
    value_field = 'id'

    def get_serializer_class(self):
        class LabelValueSerializer(BaseModelSerializer):
            label = serializers.CharField(source=self.label_field)
            value = serializers.CharField(source=self.value_field)

            class Meta:
                model = self.get_model()
                fields = ('label', 'value')

        return LabelValueSerializer

    def get(self, request, **kwargs):
        queryset = self.get_filtered_queryset()
        ser = self.get_serializer(queryset, many=True)
        return self.simple_response(ser.data)



from djackal.settings import djackal_settings
from djackal.utils import isiter


class skip:
    """
    When null is this class, that will removed in value dict.
    """
    pass


class InspectorException(Exception):
    def __init__(self, message):
        self.message = message


class InspectActionException(Exception):
    def __init__(self, key, value, action=None, param=None, message=None):
        self.action = action
        self.key = key
        self.value = value
        self.param = param
        self.message = message


class Inspector:
    """
    example of inspector map

    {
        key: {
            'required': True or False,
            'type': ['int', 'float', 'number', 'str', 'list', 'dict', 'bool'],
            'default': {DEFAULT_VALUE},
            'convert': [int, float, str, list, convert_function, ...],
        },
        another_key: {
            ...actions
        }
    }
    """
    none_values = None
    type_maps = {
        'int': int,
        'float': float,
        'number': (int, float),
        'str': str,
        'bool': bool,
        'list': (tuple, list, set),
        'dict': dict,
    }
    allow_actions = ['required', 'default', 'type', 'convert', 'unknown']

    def __init__(self, _map):
        self.inspect_map = _map

    def get_none_values(self):
        return self.none_values if self.none_values is not None else djackal_settings.DEFAULT_NONE_VALUES

    def action_unknown(self, key, value, param=None):
        """
        When data not exists in inspect_map
        """
        return skip

    def action_required(self, key, value, param):
        """
        When field has required.
        """
        required = param
        if required and value is None:
            raise InspectActionException(
                key, value, action='required', param=param,
                message='required field {} is not exists'.format(key)
            )
        return value

    def action_convert(self, key, value, param):
        """
        When field has convert, convert value
        """
        if not callable(param):
            raise InspectorException(
                message='{} convert param is not callable'.format(key)
            )
        return param(value)

    def action_type(self, key, value, param):
        """
        When field has type, check type.
        """
        if param not in self.type_maps:
            raise InspectActionException(
                key, value, action='type', param=param,
                message='type: {} is not allowed'.format(param)
            )
        if value is None:
            return value

        standard = self.type_maps[param]
        value_type = type(value)

        if (isiter(standard) and value_type in standard) or value_type is standard:
            return value
        raise InspectActionException(
            key, value, action='type', param=param,
            message='value type: {} is not {}'.format(value_type, param)
        )

    def action_default(self, key, value, param):
        """
        When field has default and value in none_values, return default
        """
        default = param
        if value is None:
            if param is skip:
                return skip
            elif callable(default):
                return default()
            return default
        return value

    def sort_field(self, field):
        if 'required' in field and 'default' in field:
            raise InspectorException(
                'Validation failed: required action cannot be used with default.'
            )

        result = {}
        for action in self.allow_actions:
            if action in field:
                result[action] = field[action]
        return result

    def inspect(self, data):
        result = {}

        # fields in inspect_map
        for key, field in self.inspect_map.items():
            raw_value = data.get(key)
            result[key] = raw_value if raw_value not in self.get_none_values() else None
            field = self.sort_field(field)
            for action, param in field.items():
                value = self.action(action, key, result[key], param)
                if value is skip:
                    result.pop(key, None)
                    break
                result[key] = value

        # data not in inspect_map
        unknown_data = {key: value for key, value in data.items() if key not in self.inspect_map}
        for key, value in unknown_data.items():
            value = self.action('unknown', key, value, None)
            if value is skip:
                continue
            result[key] = value
        return result

    def action(self, action, key, value, param):
        if action not in self.allow_actions:
            raise InspectorException('invalid action: {}'.format(action))
        method = getattr(self, 'action_{}'.format(action))
        return method(key, value, param)



class BaseFilterFunc:
    def action(self, name, value):
        func = getattr(self, name, None)
        if func is None:
            raise KeyError(f'QuanFunc not found by name: \'{name}\'')
        if not callable(func):
            raise ValueError(f'QuanFunc is not callable: \'{name}\'')

        return func(value)


class DefaultFilterFunc(BaseFilterFunc):
    def to_bool(self, value):
        if value.lower() in ['true', 't', 'yes', 'y', 'on', '1']:
            return True
        elif value.lower() in ['false', 'f', 'no', 'n', 'off', '0']:
            return False
        return None

    def to_int(self, value):
        return int(value)



from rest_framework import permissions

from djackal.exceptions import PermissionException

REAL_SAFE_METHOD = ('OPTIONS', 'HEAD')
POST = 'POST'
GET = 'GET'
PATCH = 'PATCH'
PUT = 'PUT'
DELETE = 'DELETE'


class ErraPermission(permissions.BasePermission):
    hard = False
    erra = None
    status_code = 403
    code = None
    message = None

    def get_message(self, request, view, obj=None):
        return self.message

    def get_erra(self, request, view, obj=None):
        return self.erra

    def get_status_code(self, request, view, obj=None):
        return self.status_code

    def get_code(self, request, view, obj=None):
        return self.code

    def raise_error(self, request, view, obj=None):
        raise PermissionException(
            permission=self,
            erra=self.get_erra(request, view, obj),
            code=self.get_code(request, view, obj),
            message=self.get_message(request, view, obj),
            status_code=self.get_status_code(request, view, obj)
        )

    def handle(self, request, view):
        return True

    def handle_object(self, request, view, obj):
        return True

    def has_permission(self, request, view):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        if self.handle(request, view):
            return True

        if self.hard:
            self.raise_error(request, view)
        else:
            return False

    def has_object_permission(self, request, view, obj):
        """
        Return `True` if permission is granted, `False` otherwise.
        """
        if self.handle_object(request, view, obj):
            return True

        if self.hard:
            self.raise_error(request, view, obj=obj)
        else:
            return False


class _MethodPermission(ErraPermission):
    allow_method = None

    def handle(self, request, view):
        if request.method in REAL_SAFE_METHOD:
            return True
        return request.method == self.allow_method

    def handle_object(self, request, view, obj):
        return self.handle(request, view)


class IsGet(_MethodPermission):
    """
    if request method is GET return True
    """
    allow_method = GET


class IsPost(_MethodPermission):
    """
    if request method is POST return True
    """
    allow_method = POST


class IsPut(_MethodPermission):
    """
    if request method is PUT return True
    """
    allow_method = PUT


class IsPatch(_MethodPermission):
    """
    if request method is PATCH return True
    """
    allow_method = PATCH


class IsDelete(_MethodPermission):
    """
    if request method is DELETE return True
    """
    allow_method = DELETE



from django.db.models import QuerySet
from django.test import TransactionTestCase, TestCase
from rest_framework.test import APITestCase


class _TestMixin:
    def assertLen(self, length, seq):
        if isinstance(seq, QuerySet):
            assert length == seq.count()
        assert length == len(seq)

    def assertLenEqual(self, seq1, seq2):
        assert len(seq1) == len(seq2)

    def assertSuccess(self, response):
        assert 200 == response.status_code

    def assertStatusCode(self, code, response):
        assert code == response.status_code


class DjackalAPITestCase(APITestCase, _TestMixin):
    pass


class DjackalTransactionTestCase(TransactionTestCase, _TestMixin):
    pass


class DjackalTestCase(TestCase, _TestMixin):
    pass



from importlib import import_module

from django.conf import settings
from django.test.signals import setting_changed

default_param_func = 'djackal.param_funcs.DefaultParamFunc'

DEFAULTS = {
    'PAGE_SIZE': 10,
    'MAX_PAGE_SIZE': 100,

    'DEFAULT_PAGINATION_CLASS': 'djackal.pagination.PageNumberPagination',
    'DEFAULT_INSPECTOR': 'djackal.inspector.Inspector',

    'EXCEPTION_HANDLER': 'djackal.views.handler.exception_handler',

    'DEFAULT_NONE_VALUES': ([], {}, '', None),

    'INITIALIZER': None,

    'PARAM_FUNC_CLASSES': [
        default_param_func,
    ],

    'SINGLE_APP': False,
    'SINGLE_APP_NAME': None,
}

IMPORT_STRINGS = [
    'PARAM_FUNC_CLASSES',

    'DEFAULT_PAGINATION_CLASS',
    'DEFAULT_INSPECTOR',

    'EXCEPTION_HANDLER',
]


class DjackalSettings:
    def __init__(self, custom_settings=None, defaults=None, import_strings=None):
        if custom_settings:
            self._custom_settings = custom_settings

        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self._cached_attrs = set()

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid Djackal setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.custom_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    @property
    def custom_settings(self):
        if not hasattr(self, '_custom_settings'):
            self._custom_settings = getattr(settings, 'DJACKAL', {})
        return self._custom_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_custom_settings'):
            delattr(self, '_custom_settings')


def perform_import(val, setting_name):
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    elif isinstance(val, dict):
        return {key: perform_import(item, setting_name) for key, item in val.items()}

    return val


def import_from_string(val, setting_name):
    try:
        module_path, class_name = val.rsplit('.', 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '%s' for Djackal setting '%s'. %s: %s." % (val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


djackal_settings = DjackalSettings(None, DEFAULTS, IMPORT_STRINGS)


def reload_djackal_settings(*args, **kwargs):
    setting = kwargs['setting']
    if setting == 'DJACKAL':
        djackal_settings.reload()


setting_changed.connect(reload_djackal_settings)



class BindMixin:
    bound_fields = []
    bind_field_name = 'extra'

    def __init__(self, *args, **kwargs):
        self.bind_field_keys = list(self.bound_fields)
        extra = kwargs.pop('extra', dict())
        for bound_key in self.bound_fields:
            if bound_key in kwargs:
                value = kwargs.pop(bound_key)
            else:
                if type(self.bound_fields) is dict:
                    default_value = self.bound_fields.get(bound_key, None)
                    value = default_value() if callable(default_value) else default_value
                else:
                    continue
            extra.setdefault(bound_key, value)
        kwargs['extra'] = extra
        super().__init__(*args, **kwargs)

    def __setattr__(self, key, value):
        if key in ['bound_fields', 'bind_field_name', 'bind_field_keys']:
            super().__setattr__(key, value)
            return
        if key in self.bind_field_keys:
            bind_hash = self.__bound_data
            bind_hash[key] = value
            setattr(self, self.bind_field_name, bind_hash)
            return
        else:
            super().__setattr__(key, value)
            return

    def __getattribute__(self, item):
        if item in ['bound_fields', 'bind_field_name', 'bind_field_keys']:
            return super().__getattribute__(item)
        elif item in self.bind_field_keys:
            bind_hash = self.__bound_data
            value = bind_hash.get(item)
            if value is None and type(self.bound_fields) is dict:
                default = self.bound_fields[item]
                value = default() if callable(default) else default
            return value
        else:
            return super().__getattribute__(item)

    @property
    def __bound_data(self):
        return getattr(self, self.bind_field_name, dict())
